/*
 * Copyright 2013 Canonical Ltd.
 *
 * This file is part of powerd.
 *
 * powerd is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * powerd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <gudev/gudev.h>

#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include <glib-object.h>
#include <gio/gio.h>
#include "powerd-dbus.h"
#include "powerd.h"

#include <hybris/input/input_stack_compatibility_layer.h>
#include <hybris/input/input_stack_compatibility_layer_codes_key.h>

#include <hybris/surface_flinger/surface_flinger_compatibility_layer.h>

namespace
{

enum SysPowerStates {
    SUSPEND = 0,
    ACTIVE
};

struct SysStateRequest {
    GString *name;
    GString *owner;
    SysPowerStates state;
};

enum {
    ACQUIRE_PARTIAL_WAKE_LOCK = 0,
    RELEASE_WAKE_LOCK,
    REQUEST_STATE,
    OUR_FD_COUNT
};

const char * const PATHS[] = {
    "/sys/power/wake_lock",
    "/sys/power/wake_unlock",
    "/sys/power/state"
};


G_DEFINE_TYPE (PowerdSource, powerd_source, G_TYPE_OBJECT);

#if 0
struct _PowerdSourcePrivate {
    GDBusConnection * session;
    gchar * path;
    ComCanonicalPowerd * skel;
};
#endif

#define POWERD_SOURCE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), POWERD_TYPE_SOURCE, PowerdSourcePrivate))

GSList *requests = NULL;

static const char *off_state = "mem";
static const char *on_state = "on";

static bool in_suspend = false;
static int g_initialized = 0;
static int g_fds[OUR_FD_COUNT];
static int g_error = 1;

static uint button_timer = 0;
static uint activity_timer = 0;

static time_t curtime;
static int lasttime;
struct tm *tm;

static gint saved_brightness = 255;
static gchar *brightness_path;

static int ACTIVITY_TIMEOUT = 30;

enum state {
    BUTTON_DOWN,
    BUTTON_UP,
    SHUTDOWN
};

static void powerd_source_class_init (PowerdSourceClass * klass);
static void powerd_source_init       (PowerdSource *      self);
static void powerd_source_dispose    (GObject *           object);
static void powerd_source_finalize   (GObject *           object);
static const gchar * state_to_string(int state);

static state button_state  = BUTTON_UP;

/* Instance */
static void
powerd_source_init (PowerdSource *self)
{
    self->priv = POWERD_SOURCE_GET_PRIVATE(self);
    self->priv->session = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL, NULL);
    return;
}

/* Class Init */
static void
powerd_source_class_init (PowerdSourceClass *self)
{
    GObjectClass *object_class = G_OBJECT_CLASS (self);
    g_type_class_add_private (self, sizeof (PowerdSourcePrivate));
    object_class->dispose = powerd_source_dispose;
    object_class->finalize = powerd_source_finalize;
    return;
}

/* Clean up references */
static void
powerd_source_dispose (GObject *object)
{
    PowerdSource * self = POWERD_SOURCE(object);
    if (self->priv->skel != NULL) {
        g_dbus_interface_skeleton_unexport(G_DBUS_INTERFACE_SKELETON(self->priv->skel));
        g_clear_object(&self->priv->skel);
    }
    g_clear_object(&self->priv->session);
    return;
}

/* Free memory */
static void
powerd_source_finalize (GObject *object)
{
    return;
}


void get_brightness_path (void)
{
    GUdevClient *client;
    GList *devices;
    GList *device;
    const gchar *device_type;

    // detect device
    client = g_udev_client_new (NULL);
    devices = g_udev_client_query_by_subsystem (client, "backlight");
    if (devices == NULL) {
        g_warning ("Fail to query backlight devices.");
        g_object_unref (client);
        return;
    }

    for (device = devices; device != NULL; device = device->next) {
        device_type = g_udev_device_get_sysfs_attr ((GUdevDevice *)device->data, "type");
        if ((g_strcmp0 (device_type, "firmware") == 0) ||
            (g_strcmp0 (device_type, "platform") == 0) ||
            (g_strcmp0 (device_type, "raw") == 0)) {
            brightness_path = g_strdup (g_udev_device_get_sysfs_path ((GUdevDevice *)device->data));
            g_print ("found: %s\n", brightness_path);
            break;
        }
    }
    g_print ("found2: %s\n", brightness_path);

    g_list_foreach (devices, (GFunc) g_object_unref, NULL);
    g_list_free (devices);

    g_object_unref (client);
}


void
brightness_set_value (gint value)
{
    gint fd;
    gchar *filename;
    gchar *svalue;
    gint length;

    filename = g_build_filename (brightness_path, "brightness", NULL);
    fd = open(filename, O_WRONLY);
    if (fd < 0) {
        g_warning ("Fail to set brightness.");
        g_free (filename);
        return;
    }

    svalue = g_strdup_printf ("%i", value);
    length = strlen (svalue);

    if (write (fd, svalue, length) != length) {
        g_warning ("Fail to write brightness information.");
    }

    close (fd);
    g_free (svalue);
    g_free (filename);
}

gint
brightness_get_value (void)
{
    GError *error;
    gchar *svalue;
    gint value;
    gchar *filename;

    svalue = NULL;
    error = NULL;
    filename = g_build_filename (brightness_path, "brightness", NULL);
    g_file_get_contents (filename, &svalue, NULL, &error);
    if (error) {
        g_warning ("Fail to get brightness value: %s", error->message);
        value = -1;
        g_error_free (error);
    } else {
        value = atoi (svalue);
    }

    g_free (filename);

    return value;

}

static int
open_file_descriptors(const char * const paths[])
{
    int i;
    for (i=0; i<OUR_FD_COUNT; i++) {
        int fd = open(paths[i], O_RDWR);
        if (fd < 0) {
            fprintf(stderr, "fatal error opening \"%s\"\n", paths[i]);
            g_error = errno;
            return -1;
        }
        g_fds[i] = fd;
    }

    g_error = 0;
    return 0;
}

static inline void
initialize_fds(void)
{
    if (g_initialized == 0) {
        if(open_file_descriptors(PATHS) < 0) {
            on_state = "on";
            off_state = "mem";
        }
        g_initialized = 1;
    }
}


gboolean call_shutdown(gpointer data)
{
    gchar   *argv[] = {"poweroff",
                        NULL};
    if (button_state == BUTTON_DOWN) {
        button_state = SHUTDOWN;
        brightness_set_value(0);
        sf_blank(0);
        g_spawn_async(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL);
    }
}

int set_screen_state(void)
{
    initialize_fds();

    if (g_error)
        goto failure;

    char buf[32];
    int len;
    if(in_suspend) {
        len = snprintf(buf, sizeof(buf), "%s", on_state);
        in_suspend = false;
    } else {
        if (activity_timer > 0) {
            g_source_remove(activity_timer);               
            activity_timer = 0;
        }
        saved_brightness = brightness_get_value();
        brightness_set_value(0);
        sf_blank(0);
        len = snprintf(buf, sizeof(buf), "%s", off_state);
        in_suspend = true;
    }

    buf[sizeof(buf) - 1] = '\0';
    len = write(g_fds[REQUEST_STATE], buf, len);
    if(len < 0) {
    failure:
        printf("Failed setting last user activity: g_error=%d\n", g_error);
    }

    //Need to wait for system wakeup otherwise we will get
    //permission denied if we try to ake the screen up before that.
    if (!in_suspend) {
        sleep(1);
        sf_unblank(0);
        brightness_set_value(saved_brightness);
    }

    return 0;
}


void on_new_event(Event* event, void* context)
{
    
    if (activity_timer > 0) {
        g_source_remove(activity_timer);               
        activity_timer = 0;
    }

    switch(event->type)
    {
        case KEY_EVENT_TYPE:
            if (event->details.key.key_code == ISCL_KEYCODE_POWER && button_state != SHUTDOWN) {
                if (event->action == 1) {
                    button_state = BUTTON_UP;
                    curtime = time(0);
                    tm = localtime (&curtime);
                    if (tm->tm_sec - lasttime < 2)
                        set_screen_state();
                } else if (event->action == 0) {
                    button_state = BUTTON_DOWN;
                    curtime = time(0);
                    tm = localtime (&curtime);
                    lasttime = tm->tm_sec;
                    if (button_timer > 0)
                    {
                        g_source_remove(button_timer);
                        button_timer = 0;
                    }
                    button_timer = g_timeout_add_seconds(2, call_shutdown, NULL);
                }
            }
            break;
        default:
            break;
    }
}

/* simple helper for debugging */
static const gchar *
state_to_string(int state)
{
    switch (state) {
        case ACTIVE:
            return "STATE_ACTIVE";
            break;
        // you really shouldnt get this state since it's not requestable
        // externally
        case SUSPEND:
            return "STATE_SUSPEND";
            break;
        default:
            return "UNKNOWN";
            break;
    }
}

static gboolean
handle_list_requests (PowerdSource *obj, GDBusMethodInvocation *invocation)
{
    GSList *iterator = NULL;
    if (!requests) {
        g_message("no requests active");
    }
    else {
        for (iterator = requests; iterator; iterator = iterator->next) {
            SysStateRequest *sr = (SysStateRequest *)iterator->data;
            g_message("Owner: %s, Name: %s, State: %d (%s)",\
                sr->owner->str, sr->name->str, sr->state,\
                state_to_string(sr->state));
        }
    }
    g_dbus_method_invocation_return_value(invocation, NULL);
    return TRUE;
}

}

int main(int argc, char** argv)
{
    GError *error = NULL;
    GMainLoop* main_loop = NULL;

    main_loop = g_main_loop_new (NULL, FALSE);
    g_type_init();
    PowerdSource *source = (PowerdSource *)g_object_new(POWERD_TYPE_SOURCE, NULL);
    source->priv->skel = com_canonical_powerd_skeleton_new();
    source->priv->path = "/com/canonical/powerd";

    AndroidEventListener listener;
    listener.on_new_event = on_new_event;
    listener.context = NULL;

    InputStackConfiguration config = {
    enable_touch_point_visualization : false,
    default_layer_for_touch_point_visualization : 10000
    };

    android_input_stack_initialize(&listener, &config);
    android_input_stack_start();

    get_brightness_path();

    if (!g_dbus_interface_skeleton_export(G_DBUS_INTERFACE_SKELETON(source->priv->skel),
            source->priv->session, source->priv->path, &error)) {
        g_error("Unable to export skeleton on path '%s': %s", source->priv->path, error->message);
        g_error_free(error);
        error = NULL;
        return -1;
    }

//    g_signal_connect(G_OBJECT(source->priv->skel), "handle-request-state",\
        G_CALLBACK(handle_request_state), source);
    g_signal_connect(G_OBJECT(source->priv->skel), "handle-list-requests",\
        G_CALLBACK(handle_list_requests), source);

    g_main_loop_run(main_loop);

    android_input_stack_stop();
    android_input_stack_shutdown();

    // XXX - mfisch need to cleanup our list of active requests
    return 0;
}
